import logging
from typing import Callable, Iterable, Optional, Type

from .monitor import Monitor, StackMonitor


class StackedFormatter(logging.Formatter):
    def __init__(self, parent_formatter: logging.Formatter,
                 child_formatter: logging.Formatter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent_formatter = parent_formatter
        self._child_formatter = child_formatter

    def format(self, record: logging.LogRecord) -> str:
        record.msg = self._child_formatter.format(record)
        return self._parent_formatter.format(record)


class IndentingFormatter(logging.Formatter):
    def __init__(self, monitor: Monitor, indentation: str,
                 deepest_indentation: str, max_level: Optional[int], *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._monitor = monitor
        self._indentation = indentation
        self._deepest_indentation = deepest_indentation
        self._max_level = max_level

    def format(self, record: logging.LogRecord) -> str:
        message = record.getMessage()
        level = self._monitor.get_level()

        if self._max_level and level > self._max_level:
            return self._indentation * (
                self._max_level - 1) + self._deepest_indentation + message
        else:
            return self._indentation * level + message


class DefaultLogger:
    """Used just to improve rendering of API documentation."""
    value = (logging.Logger, )

    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return '(logging.Logger,)'

    def __iter__(self):
        pass


def install(
    indentation: str = '  ',
    deepest_indentation: str = '..',
    max_level: Optional[int] = 5,
    handlers: Optional[Iterable[logging.Handler]] = None,
    logger: Optional[logging.Logger] = None,
    logging_classes: Iterable[Type] = DefaultLogger(),
    logging_functions: Iterable[Callable] = ()
) -> None:
    """Wraps formatters of each handler of the `logger` with the special
    `IndentingFormatter`.

    If `logger` is `None` (default), root logger will be used. If `logger`
    doesn't have any handlers added, an instance of `logging.StreamHandler`
    will be created and added.

    Handlers to be wrapped can be also provided through `handlers` argument.
    In this case `loigger` argument is not taken into consideration.

    If your application writes logs by calling custom logging functions, these
    functions should be registered in `logging_functions`. Alternatively, if
    they are methods of a class, this class can be registered in
    `logging_classes`. There is no need of registering classes that derive from
    `logging.Logger`.

    Indentation is created as `indentation * indentation_level`, where
    `indentation_level` is calculated from current location in the call stack.
    `indentation_level` can be limited by `max_level`, or unlimited when
    `max_level` is `None`. Whenever `indentation_level` exceeds the limit,
    indentation is created as
    `indentation * (max_level - 1) + deepest_indentation`.

    Arguments:
        indentation: String concatenated with the message to create a
            single-level indentation.
        deepest_indentation: String preceeding the message in case when
            indentation level exceeds the limit.
        max_level: Limit of the indentation level. Use `None` for disabling the
            limit.
        handlers: Logging handlers, formatters of which are wrapped by
            `IndentingFormatter`.
        logger: Logger, handlers of which are explored to have their formatters
            wrapped.
        logging_classes: Collection of classes, methods of which are considered
            as logging functions.
        logging_functions: Collection of logging functions.
    """

    if isinstance(logging_classes, DefaultLogger):
        logging_classes = DefaultLogger.value

    monitor = StackMonitor(logging_classes, logging_functions)

    if handlers is None:
        if logger is None:
            logger = logging.getLogger()
        if not logger.handlers:
            # Default handler (`logging.lastResort`) could be used here,
            # but it is not part of the official API
            logger.addHandler(logging.StreamHandler())
        handlers = logger.handlers

    for handler in handlers:
        ind_fmt = IndentingFormatter(monitor, indentation, deepest_indentation,
                                     max_level)
        if handler.formatter is None:
            handler.setFormatter(ind_fmt)
        else:
            stk_fmt = StackedFormatter(handler.formatter, ind_fmt)
            handler.setFormatter(stk_fmt)



import gc
import inspect
from typing import Callable


def func_in_frame_info(frame_info: inspect.FrameInfo) -> Callable:
    """Find callable corresponding to given frame_info."""
    f_code = frame_info.frame.f_code
    for obj in gc.get_referrers(f_code):
        if hasattr(obj, '__code__') and obj.__code__ is f_code:
            return obj
    raise RuntimeError("Callable not found")



import inspect
import operator
from abc import ABC, abstractmethod
from itertools import takewhile
from typing import Callable, Collection, Iterable, List, Type

from .utils import func_in_frame_info


class Monitor(ABC):
    @abstractmethod
    def get_level(self) -> int:
        pass


class StackMonitor(Monitor):
    def __init__(self,
                 logging_classes: Iterable[Type] = (),
                 logging_functions: Iterable[Callable] = ()):
        self._last_identifiers: Collection = tuple()
        self._identifier_groups: List = []
        logger_methods = [
            m for c in logging_classes for m in vars(c).values()
            if inspect.isfunction(m)
        ]
        self._logging_functions = (*logger_methods, *logging_functions)

    def get_level(self) -> int:
        reversed_stack = reversed(inspect.stack())
        user_stack = takewhile(self._corresponds_to_user_functon,
                               reversed_stack)
        identifiers = tuple(
            id(frame_info.frame.f_code) for frame_info in user_stack)
        return self._get_level(identifiers)

    def _corresponds_to_user_functon(self,
                                     frame_info: inspect.FrameInfo) -> bool:
        return func_in_frame_info(frame_info) not in self._logging_functions

    def _get_level(self, identifiers: Collection) -> int:

        common_part = takewhile(lambda pair: operator.eq(*pair),
                                zip(self._last_identifiers, identifiers))
        common_len = sum(1 for _ in common_part)
        identifiers_len = len(identifiers)
        last_identifiers_len = len(self._last_identifiers)

        if common_len == last_identifiers_len:
            if identifiers_len > last_identifiers_len:  # [Next level]
                self._identifier_groups.append(identifiers_len)
            # else: [Same level]

        elif common_len < last_identifiers_len:
            if common_len < identifiers_len:  # [Jump]
                selector = lambda x: x <= common_len  # noqa: E731
            else:  # [Previous level]
                selector = lambda x: x < common_len  # noqa: E731

            self._identifier_groups = list(
                takewhile(selector, self._identifier_groups))
            self._identifier_groups.append(identifiers_len)

        self._last_identifiers = identifiers
        level = len(self._identifier_groups) - 1

        return level



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['indentedlogs']

package_data = \
{'': ['*']}

install_requires = \
['bandit>=1.6.2,<2.0.0', 'pylint>=2.4.4,<3.0.0', 'radon>=4.1.0,<5.0.0']

setup_kwargs = {
    'name': 'indentedlogs',
    'version': '0.1.1',
    'description': 'Adds indentation to the log messages according to their location in the call stack.',
    'long_description': "# indentedlogs\n\nAdds indentation to the log messages according to their location in the call stack.\n\n* Documentation: <https://gergelyk.github.io/python-indentedlogs>\n* Repository: <https://github.com/gergelyk/python-indentedlogs>\n* Package: <https://pypi.python.org/pypi/indentedlogs>\n* Author: [grzegorz.krason@gmail.com](mailto:grzegorz.krason@gmail.com)\n* License: [MIT](LICENSE)\n\n## Requirements\n\nThis package requires CPython 3.8 or compatible. If you have other version already installed, you can switch using `pyenv`. It must be installed as described in the [manual](https://github.com/pyenv/pyenv).\n\n```sh\npyenv install 3.8.2\npyenv local 3.8.2\n```\n\n## Installation\n\n```sh\npip install indentedlogs\n```\n\n## Examples\n\n```sh\npython examples/demo0.py\npython examples/demo1.py\npython examples/demo2.py\n```\n\n## Development\n\n```sh\n# Preparing environment\npip install --user poetry  # unless already installed\npoetry install\n\n# Auto-formatting\npoetry run docformatter -ri indentedlogs tests\npoetry run isort -rc indentedlogs tests\npoetry run yapf -r -i indentedlogs tests\n\n# Checking coding style\npoetry run flake8 indentedlogs tests\n\n# Checking composition and quality\npoetry run mypy indentedlogs tests\npoetry run pylint indentedlogs tests\npoetry run bandit indentedlogs tests\npoetry run radon cc indentedlogs tests\npoetry run radon mi indentedlogs tests\n\n# Testing with coverage\npoetry run pytest --cov indentedlogs --cov tests\n\n# Rendering documentation\npoetry run mkdocs serve\n\n# Building package\npoetry build\n\n# Releasing\npoetry version minor  # increment selected component\ngit tag ${$(poetry version)[2]}\ngit push --tags\npoetry build\npoetry publish\npoetry run mkdocs build\npoetry run mkdocs gh-deploy -b gh-pages\n```\n\n## Donations\n\nIf you find this software useful and you would like to repay author's efforts you are welcome to use following button:\n\n[![Donate](https://www.paypalobjects.com/en_US/PL/i/btn/btn_donateCC_LG.gif)](https://www.paypal.com/cgi-bin/webscr?cmd=_s-xclick&hosted_button_id=D9KUJD9LTKJY8&source=url)\n",
    'author': 'Grzegorz Krasoń',
    'author_email': 'grzegorz.krason@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gergelyk.github.io/python-indentedlogs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
